# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Provides support for basic commands of G90 alarm panel.
"""
from __future__ import annotations
import logging
import json
import asyncio
from asyncio import Future
from asyncio.protocols import DatagramProtocol
from asyncio.transports import DatagramTransport, BaseTransport
from typing import Optional, Tuple, List, Any
from dataclasses import dataclass
from ..exceptions import (G90Error, G90TimeoutError)
from ..const import G90Commands


_LOGGER = logging.getLogger(__name__)
G90BaseCommandData = List[Any]


@dataclass
class G90Header:
    """
    Represents JSON structure of the header used in alarm panel commands.

    :meta private:
    """
    code: Optional[int] = None
    data: Optional[G90BaseCommandData] = None


class G90BaseCommand(DatagramProtocol):
    """
    Implements basic command handling for alarm panel protocol.
    """
    # pylint: disable=too-many-instance-attributes
    # Lock need to be shared across all of the class instances
    _sk_lock = asyncio.Lock()

    def __init__(self, host: str, port: int, code: G90Commands,
                 data: Optional[G90BaseCommandData] = None,
                 local_port: Optional[int] = None,
                 timeout: float = 3.0, retries: int = 3) -> None:
        # pylint: disable=too-many-arguments
        self._remote_host = host
        self._remote_port = port
        self._local_port = local_port
        self._code = code
        self._timeout = timeout
        self._retries = retries
        self._data = '""'
        self._result: G90BaseCommandData = []
        self._connection_result: Optional[
            Future[Tuple[str, int, bytes]]
        ] = None
        if data:
            self._data = json.dumps([code, data],
                                    # No newlines to be inserted
                                    indent=None,
                                    # No whitespace around entities
                                    separators=(',', ':'))
        self._resp = G90Header()

    # Implementation of datagram protocol,
    # https://docs.python.org/3/library/asyncio-protocol.html#datagram-protocols
    def connection_made(self, transport: BaseTransport) -> None:
        """
        Invoked when connection is established.
        """

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """
        Invoked when connection is lost.
        """

    def datagram_received(self, data: bytes, addr: Tuple[str, int]) -> None:
        """
        Invoked when datagram is received.
        """
        if asyncio.isfuture(self._connection_result):
            if self._connection_result.done():
                _LOGGER.warning('Excessive packet received'
                                ' from %s:%s: %s',
                                addr[0], addr[1], data)
                return
            self._connection_result.set_result((*addr, data))

    def error_received(self, exc: Exception) -> None:
        """
        Invoked when error is received.
        """
        if (
            asyncio.isfuture(self._connection_result) and not
            self._connection_result.done()
        ):
            self._connection_result.set_exception(exc)

    async def _create_connection(self) -> (
        Tuple[DatagramTransport, DatagramProtocol]
    ):
        """
        Creates UDP connection to the alarm panel.
        """
        loop = asyncio.get_running_loop()

        _LOGGER.debug('Creating UDP endpoint for %s:%s',
                      self.host, self.port)
        local_addr = None
        if self._local_port:
            local_addr = ('0.0.0.0', self._local_port)

        transport, protocol = await loop.create_datagram_endpoint(
            lambda: self,
            remote_addr=(self.host, self.port),
            allow_broadcast=True,
            local_addr=local_addr)

        return (transport, protocol)

    def to_wire(self) -> bytes:
        """
        Returns the command in wire format.
        """
        wire = bytes(f'ISTART[{self._code},{self._code},{self._data}]IEND\0',
                     'utf-8')
        _LOGGER.debug('Encoded to wire format %s', wire)
        return wire

    def from_wire(self, data: bytes) -> G90BaseCommandData:
        """
        Parses the response from the alarm panel.
        """
        _LOGGER.debug('To be decoded from wire format %s', data)
        try:
            self._parse(data.decode('utf-8'))
        except UnicodeDecodeError as exc:
            raise G90Error(
                'Unable to decode response from UTF-8'
            ) from exc
        return self._resp.data or []

    def _parse(self, data: str) -> None:
        """
        Processes the response from the alarm panel.
        """
        if not data.startswith('ISTART'):
            raise G90Error('Missing start marker in data')
        if not data.endswith('IEND\0'):
            raise G90Error('Missing end marker in data')
        payload = data[6:-5]
        _LOGGER.debug("Decoded from wire: JSON string '%s'", payload)

        resp = None
        if payload:
            try:
                resp = json.loads(payload)
            except json.JSONDecodeError as exc:
                raise G90Error('Unable to parse response as JSON:'
                               f" '{payload}'") from exc

            if not isinstance(resp, list):
                raise G90Error('Mailformed response,'
                               f" 'list' expected: '{payload}'")

        if resp is not None:
            self._resp = G90Header(*resp)
            _LOGGER.debug('Parsed from wire: %s', self._resp)

            if not self._resp.code:
                raise G90Error(f"Missing code in response: '{payload}'")
            # Check there is data if the response is non-empty
            if not self._resp.data:
                raise G90Error(f"Missing data in response: '{payload}'")

            if self._resp.code != self._code:
                raise G90Error(
                    'Wrong response - received code '
                    f"{self._resp.code}, expected code {self._code}")

    @property
    def result(self) -> G90BaseCommandData:
        """
        The result of the command.
        """
        return self._result

    @property
    def host(self) -> str:
        """
        The hostname/IP address of the alarm panel.
        """
        return self._remote_host

    @property
    def port(self) -> int:
        """
        The port of the alarm panel.
        """
        return self._remote_port

    async def process(self) -> G90BaseCommand:
        """
        Processes the command.
        """
        # Disallow using `NONE` command, which is intended to use by inheriting
        # classes overriding `process()` method
        if self._code == G90Commands.NONE:
            raise G90Error("'NONE' command code is disallowed")

        transport, _ = await self._create_connection()
        attempts = self._retries
        while True:
            attempts = attempts - 1
            loop = asyncio.get_running_loop()
            self._connection_result = loop.create_future()
            async with self._sk_lock:
                _LOGGER.debug('(code %s) Sending request to %s:%s',
                              self._code, self.host, self.port)
                transport.sendto(self.to_wire())
                done, _ = await asyncio.wait([self._connection_result],
                                             timeout=self._timeout)
            if self._connection_result in done:
                break
            # Cancel the future to signal protocol handler it is no longer
            # valid, the future will be re-created on next retry
            self._connection_result.cancel()
            if not attempts:
                transport.close()
                raise G90TimeoutError()
            _LOGGER.debug('Timed out, retrying')
        transport.close()
        (host, port, data) = self._connection_result.result()
        _LOGGER.debug('Received response from %s:%s', host, port)
        if self.host != '255.255.255.255':
            if self.host != host or host == '255.255.255.255':
                raise G90Error(f'Received response from wrong host {host},'
                               f' expected from {self.host}')
        if self.port != port:
            raise G90Error(f'Received response from wrong port {port},'
                           f' expected from {self.port}')

        ret = self.from_wire(data)
        self._result = ret
        return self

    def __repr__(self) -> str:
        """
        Returns string representation of the command.
        """
        return f'Command: {self._code}, request: {self._data},' \
            f' response: {self._resp.data}'



# Copyright (c) 2023 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Performs runtime configuration and exposes custom fixtures for Pytest.
"""
from __future__ import annotations
from typing import AsyncIterator, Callable, Any
import asyncio
from unittest.mock import MagicMock, DEFAULT
import pytest
from pyg90alarm.notifications.base import G90NotificationsBase
from .device_mock import DeviceMock


def pytest_configure(config: pytest.Config) -> None:
    """
    Configures `pytest`.
    """
    config.addinivalue_line("markers", "g90device")


@pytest.fixture
async def mock_device(
    request: pytest.FixtureRequest,
    unused_udp_port_factory: Callable[..., int],
    unused_tcp_port_factory: Callable[..., int]
) -> AsyncIterator[DeviceMock]:
    """
    Fixture to instantiate a simulated G90 device allocating random unused
    ports for network exchanges.

    The fixture should be customized with `g90device` mark containing
    `sent_data` and `notification_data` lists of bytes to simulate the messages
    the device sends back to client and notification messages sent to client,
    respectively.
    """
    marker = getattr(
        request.node
        .get_closest_marker('g90device'),
        'kwargs', {}
    )
    data = marker.get('sent_data', [])
    notification_data = marker.get('notification_data', [])
    cloud_notification_data = marker.get('cloud_notification_data', [])
    cloud_upstream_data = marker.get('cloud_upstream_data', [])

    # Allocate unused ports to listen for client requests on, and to send
    # notification messages to, respectively

    # Note the `unised_udp_port_factory` comes from `pytest-asyncio` package
    device_port = unused_udp_port_factory()
    notification_port = unused_udp_port_factory()
    proxy_port = unused_tcp_port_factory()
    upstream_port = unused_tcp_port_factory()
    device = DeviceMock(
        data, notification_data,
        device_port=device_port, notification_port=notification_port,
        cloud_port=proxy_port,
        cloud_notification_data=cloud_notification_data,
        cloud_upstream_port=upstream_port,
        cloud_upstream_data=cloud_upstream_data
    )
    await device.start()
    yield device
    await device.stop()


def data_receive_awaitable(obj: G90NotificationsBase) -> asyncio.Future[bool]:
    """
    Creates an awaitable future that resolves when data is received by
    notification handler.

    This function wraps either the data_received or datagram_received method of
    the provided notification handler object with a mock that will resolve the
    returned future when called.

    :param obj: The notification handler object whose receive method will be
      wrapped
    :return: Future that resolves when data is received
    :raises ValueError: If no suitable method to wrap is found on the object
    """
    method = None
    method_name = None

    if hasattr(obj, 'data_received'):
        method = obj.data_received
        method_name = 'data_received'
    if hasattr(obj, 'datagram_received'):
        method = obj.datagram_received
        method_name = 'datagram_received'
    if not method or not method_name:
        raise ValueError('No method found to wrap')

    future = asyncio.get_running_loop().create_future()

    def wrapper(*_args: Any) -> DEFAULT:
        """
        Wrapper function that sets the future's result when called.

        :param _args: Arguments passed to the original method (unused)
        :return: Return value from unittest.mock.DEFAULT
        """
        future.set_result(True)
        # Signal that the original method should be called
        return DEFAULT

    setattr(obj, method_name, MagicMock(wraps=method, side_effect=wrapper))
    return future



"""
Tests for the G90AlertConfig class.
"""
from __future__ import annotations
import pytest
from pyg90alarm.alarm import (
    G90Alarm,
)
from pyg90alarm.local.config import (
    G90AlertConfigFlags,
)
from .device_mock import DeviceMock


@pytest.mark.g90device(sent_data=[
    b'ISTART[117,[1]]IEND\0',
])
async def test_alert_config(mock_device: DeviceMock) -> None:
    """
    Tests for retrieving alert configuration from the device.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    config = await g90.alert_config.flags
    assert await mock_device.recv_data == [b'ISTART[117,117,""]IEND\0']
    assert isinstance(config, G90AlertConfigFlags)
    assert G90AlertConfigFlags.AC_POWER_FAILURE in config


@pytest.mark.g90device(sent_data=[
    b"ISTART[117,[1]]IEND\0",
    b"ISTART[117,[3]]IEND\0",
    b"ISTARTIEND\0",
])
async def test_set_alert_config(mock_device: DeviceMock) -> None:
    """
    Tests for setting alert configuration to the the device.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    await g90.alert_config.set_flag(G90AlertConfigFlags.AC_POWER_FAILURE, True)
    await g90.alert_config.set_flag(G90AlertConfigFlags.HOST_LOW_VOLTAGE, True)
    assert await mock_device.recv_data == [
        b'ISTART[117,117,""]IEND\0',
        b'ISTART[117,117,""]IEND\0',
        b"ISTART[116,116,[116,[9]]]IEND\0",
    ]
    # Validate we retrieve same alert configuration just has been set
    assert await g90.alert_config.get_flag(
        G90AlertConfigFlags.AC_POWER_FAILURE) is True
    assert await g90.alert_config.get_flag(
        G90AlertConfigFlags.HOST_LOW_VOLTAGE) is True



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Extends paginated command for G90 alarm panel providing convenience interface
to work with results of paginated commands.
"""

import logging
from typing import Any, Optional, AsyncGenerator, Iterable, cast
from dataclasses import dataclass
from .paginated_cmd import G90PaginatedCommand
from ..const import (
    G90Commands,
    CMD_PAGE_SIZE,
)

_LOGGER = logging.getLogger(__name__)


@dataclass
class G90PaginatedResponse:
    """
    Response yielded from the :meth:`.G90PaginatedResult.process` method
    """
    proto_idx: int
    data: str


class G90PaginatedResult:
    """
    Processes paginated response from G90 corresponding panel commands.
    """
    # pylint: disable=too-few-public-methods
    def __init__(
        self, host: str, port: int, code: G90Commands, start: int = 1,
        end: Optional[int] = None, **kwargs: Any
    ):
        # pylint: disable=too-many-arguments
        self._host = host
        self._port = port
        self._code = code
        self._start = start
        self._end = end
        self._kwargs = kwargs

    async def process(self) -> AsyncGenerator[G90PaginatedResponse, None]:
        """
        Process paginated response yielding :class:`.G90PaginatedResponse`
        instance for each element.
        """
        page = CMD_PAGE_SIZE
        start = self._start
        count = 0
        while True:
            # The start record number is one-based, so subtract one when
            # calculating the number of the end record for the current
            # iteration
            end = start + page - 1
            # Use the smallest of requested end record number and calculated
            # one (based of page size), allows for number of records less than
            # in page
            if self._end:
                end = min(end, self._end)

            _LOGGER.debug('Invoking paginated command for %s..%s range',
                          start, end)
            cmd = await G90PaginatedCommand(
                host=self._host, port=self._port, code=self._code,
                start=start, end=end,
                **self._kwargs
            ).process()

            # The caller didn't supply the end record number, use the records
            # total since it is now known
            if not self._end:
                self._end = cmd.total
            # The supplied end record number is higher than total records
            # available, reset to the latter
            if self._end > cmd.total:
                _LOGGER.warning('Requested record range (%i) exceeds number of'
                                ' available records (%i), setting to the'
                                ' latter', self._end, cmd.total)
                self._end = cmd.total

            _LOGGER.debug('Retrieved %i records in the iteration,'
                          ' %i available in total, target end'
                          ' record number is %i',
                          cmd.count, cmd.total, self._end)

            # Produce the resulting records for the consumer
            for idx, data in enumerate(cast(Iterable[str], cmd.result)):
                # Protocol uses one-based indexes, `start` implies that so no
                # further additions to resulting value is needed.
                # Note the index provided here is running one across multiple
                # pages hence use of `start` variable
                yield G90PaginatedResponse(start + idx, data)

            # Count the number of processed records
            count += cmd.count

            # End the loop if we processed same number of sensors as in the
            # pagination header (or attempted to process more than that by
            # an error), or no records have been received
            if not cmd.count:
                break
            if cmd.start + cmd.count - 1 >= self._end:
                break
            # Move to the next page for another iteration
            start = start + page

        _LOGGER.debug('Total number of paginated entries:'
                      ' processed %s, expected %s',
                      count,
                      # Again, both end and start record numbers are one-based,
                      # so need to add one to calculate how many records have
                      # been requested
                      self._end - self._start + 1)



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Implements callbacks.
"""
from __future__ import annotations
import asyncio
from functools import (partial, wraps)
from asyncio import Task
from typing import Any, Callable, Coroutine, cast, Optional, Union
import logging

_LOGGER = logging.getLogger(__name__)

Callback = Optional[
    Union[
        Callable[..., None],
        Callable[..., Coroutine[None, None, None]],
    ]
]


class G90Callback:
    """
    Implements callbacks.
    """
    @staticmethod
    def invoke(
        callback: Callback, *args: Any, **kwargs: Any
    ) -> None:
        """
        Invokes the callback.
        """
        if not callback:
            return

        _LOGGER.debug('Attempting to invoke callback %s'
                      ' (args: %s, kwargs: %s)',
                      callback, args, kwargs)

        if not asyncio.iscoroutinefunction(callback):
            def async_wrapper(
                func: Callable[..., None]
            ) -> Callable[..., Coroutine[Any, Any, None]]:
                """
                Wraps the regular callback function into coroutine, so it could
                later be created as async task.
                """
                @wraps(func)
                async def wrapper(
                    *args: Any, **kwds: Any
                ) -> None:
                    return func(*args, **kwds)

                return cast(Callable[..., Coroutine[Any, Any, None]], wrapper)

            task = asyncio.create_task(
                async_wrapper(
                    cast(Callable[..., None], callback)
                )(*args, **kwargs)
            )
        else:
            task = asyncio.create_task(callback(*args, **kwargs))

        def reap_callback_exception(task: Task[Any]) -> None:
            """
            Reaps an exception (if any) from the task logging it, to prevent
            `asyncio` reporting that task exception was never retrieved.
            """
            exc = task.exception()
            if exc:
                _LOGGER.error(
                    "Got exception when invoking callback '%s(...)':",
                    cast(
                        Coroutine[Any, Any, None], task.get_coro()
                    ).__qualname__,
                    exc_info=exc, stack_info=False
                )

        task.add_done_callback(reap_callback_exception)

    @staticmethod
    def invoke_delayed(
        delay: float, callback: Callable[..., None], *args: Any, **kwargs: Any
    ) -> None:
        """
        Invokes the callback after a delay.
        """
        loop = asyncio.get_running_loop()
        loop.call_later(delay, partial(callback, *args, **kwargs))



# Copyright (c) 2025 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Device list.
"""
from typing import AsyncGenerator
from .device import G90Device
from .base_list import G90BaseList
from ..const import G90Commands


class G90DeviceList(G90BaseList[G90Device]):
    """
    Device list class.
    """
    async def _fetch(self) -> AsyncGenerator[G90Device, None]:
        """
        Fetch the list of devices from the panel.

        :yields: G90Device: Device entity.
        """
        devices = self._parent.paginated_result(
            G90Commands.GETDEVICELIST
        )

        async for device in devices:
            obj = G90Device(
                *device.data, parent=self._parent, subindex=0,
                proto_idx=device.proto_idx
            )

            yield obj

            # Multi-node devices (first node has already been handled
            # above)
            for node in range(1, obj.node_count):
                obj = G90Device(
                    *device.data, parent=self._parent,
                    subindex=node, proto_idx=device.proto_idx
                )
                yield obj



# Copyright (c) 2025 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Sensor list.
"""
from typing import (
    AsyncGenerator
)

from .sensor import G90Sensor
from .base_list import G90BaseList
from ..const import G90Commands


class G90SensorList(G90BaseList[G90Sensor]):
    """
    Sensor list class.
    """
    async def _fetch(self) -> AsyncGenerator[G90Sensor, None]:
        """
        Fetch the list of sensors from the panel.

        :yields: G90Sensor: Sensor entity.
        """
        entities = self._parent.paginated_result(
            G90Commands.GETSENSORLIST
        )

        async for entity in entities:
            yield G90Sensor(
                *entity.data, parent=self._parent, subindex=0,
                proto_idx=entity.proto_idx
            )



"""
Tests for G90Discovery class
"""
import pytest
from pytest import LogCaptureFixture
from pyg90alarm.local.discovery import (
    G90Discovery,
)
from pyg90alarm.local.targeted_discovery import (
    G90TargetedDiscovery,
)
from pyg90alarm.const import (
    LOCAL_TARGETED_DISCOVERY_PORT,
)

from .device_mock import DeviceMock


@pytest.mark.g90device(sent_data=[
    b'ISTART[206,["DUMMYGUID1","","","","","",0,0,0,0,"",0,0]]IEND\0',
    b'ISTART[206,["DUMMYGUID2","","","","","",0,0,0,0,"",0,0]]IEND\0',
])
async def test_discovery(mock_device: DeviceMock) -> None:
    """
    Verifies that discovery process can find devices.
    """
    g90 = G90Discovery(host=mock_device.host,
                       port=mock_device.port,
                       timeout=0.1)
    cmd = await g90.process()
    discovered = cmd.devices
    assert discovered[0].guid == 'DUMMYGUID1'
    assert discovered[0].host == mock_device.host
    assert discovered[0].port == mock_device.port
    assert await mock_device.recv_data == [b'ISTART[206,206,""]IEND\0']


@pytest.mark.g90device(sent_data=[
    b'IWTAC_PROBE_DEVICE_ACK,TSV018-3SIA'
    b',1.2,1.1,206,1.8,3,3,1,0,2,50,100\0',
])
async def test_targeted_discovery(mock_device: DeviceMock) -> None:
    """
    Verifies that discovery process can find device being targeted directly.
    """
    g90 = G90TargetedDiscovery(
        device_id='DUMMYGUID',
        host=mock_device.host,
        port=mock_device.port,
        local_port=LOCAL_TARGETED_DISCOVERY_PORT,
        timeout=0.1)
    cmd = await g90.process()
    discovered = cmd.devices
    assert discovered[0].guid == 'DUMMYGUID'
    assert discovered[0].host == mock_device.host
    assert discovered[0].port == mock_device.port
    assert await mock_device.recv_data == [b'IWTAC_PROBE_DEVICE,DUMMYGUID\0']


@pytest.mark.g90device(sent_data=[
    b'\xdeadbeef'
])
async def test_targeted_discovery_invalid_utf_response(
    mock_device: DeviceMock, caplog: LogCaptureFixture
) -> None:
    """
    Verifies that invalid UTF-8 data in response to targeted discovery is
    logged but ignored.
    """
    g90 = G90TargetedDiscovery(
        device_id='DUMMYGUID',
        host=mock_device.host,
        port=mock_device.port,
        local_port=LOCAL_TARGETED_DISCOVERY_PORT,
        timeout=0.1)

    caplog.set_level('WARNING')
    await g90.process()
    assert ''.join(caplog.messages) == (
        'Got exception, ignoring:'
        ' Unable to decode discovery response from UTF-8'
    )


@pytest.mark.g90device(sent_data=[
    b'IWTAC_PROBE_DEVICE_ACK_BAD,TSV018-3SIA'
    b',1.2,1.1,206,1.8,3,3,1,0,2,50,100\0',
])
async def test_targeted_discovery_wrong_response_start_marker(
    mock_device: DeviceMock, caplog: LogCaptureFixture
) -> None:
    """
    Verifies that wrong start marker in response to targeted discovery is
    logged but ignored.
    """
    g90 = G90TargetedDiscovery(
        device_id='DUMMYGUID',
        host=mock_device.host,
        port=mock_device.port,
        local_port=LOCAL_TARGETED_DISCOVERY_PORT,
        timeout=0.1)

    caplog.set_level('WARNING')
    await g90.process()
    assert ''.join(caplog.messages) == (
        'Got exception, ignoring: Invalid discovery response'
    )


@pytest.mark.g90device(sent_data=[
    b'IWTAC_PROBE_DEVICE_ACK,TSV018-3SIA'
    b',1.2,1.1,206,1.8,3,3,1,0,2,50,100'
])
async def test_targeted_discovery_wrong_response_end_marker(
    mock_device: DeviceMock, caplog: LogCaptureFixture
) -> None:
    """
    Verifies that wrong end marker in response to targeted discovery is logged
    but ignored.
    """
    g90 = G90TargetedDiscovery(
        device_id='DUMMYGUID',
        host=mock_device.host,
        port=mock_device.port,
        local_port=LOCAL_TARGETED_DISCOVERY_PORT,
        timeout=0.1)

    caplog.set_level('WARNING')
    await g90.process()
    assert ''.join(caplog.messages) == (
        'Got exception, ignoring: Invalid discovery response'
    )



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Python package to control G90-based alarm systems.
"""

from .alarm import G90Alarm
from .local.base_cmd import G90BaseCommand
from .local.paginated_result import G90PaginatedResult
from .notifications.base import (
    G90DeviceAlert,
)
from .entities.sensor import (
    G90Sensor, G90SensorTypes, G90SensorAlertModes, G90SensorUserFlags
)
from .entities.device import G90Device
from .local.host_info import (
    G90HostInfo, G90HostInfoWifiStatus, G90HostInfoGsmStatus
)
from .local.host_status import G90HostStatus
from .const import (
    G90MessageTypes,
    G90NotificationTypes,
    G90ArmDisarmTypes,
    G90AlertTypes,
    G90AlertSources,
    G90AlertStates,
    G90AlertStateChangeTypes,
    G90HistoryStates,
)
from .exceptions import G90Error, G90TimeoutError

__all__ = [
    'G90Alarm', 'G90BaseCommand', 'G90PaginatedResult', 'G90DeviceAlert',
    'G90Sensor', 'G90SensorTypes', 'G90SensorAlertModes', 'G90SensorUserFlags',
    'G90Device',
    'G90HostInfo', 'G90HostInfoWifiStatus', 'G90HostInfoGsmStatus',
    'G90HostStatus',
    'G90MessageTypes', 'G90NotificationTypes', 'G90ArmDisarmTypes',
    'G90AlertTypes', 'G90AlertSources', 'G90AlertStates',
    'G90AlertStateChangeTypes', 'G90HistoryStates', 'G90Error',
    'G90TimeoutError',
]



# Copyright (c) 2025 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Protocol implementation for G90 cloud communication.

This module defines the base classes and structures used for encoding and
decoding messages that flow between G90 alarm devices and their cloud servers.

Combination of Python `dataclasses` and `struct` is used to define the
protocol messages. The `dataclass` decorator is used to define the message
classes, while the `struct` module is used to define the binary format of
the messages.
"""
from __future__ import annotations
from typing import ClassVar, cast, List, Type, TypeVar, Optional
from struct import unpack, calcsize, pack, error as StructError
from dataclasses import dataclass, astuple, asdict, InitVar
from datetime import datetime, timezone
import logging

from .const import G90CloudDirection, G90CloudCommand
from ..const import G90AlertTypes

_LOGGER = logging.getLogger(__name__)
CloudBaseT = TypeVar('CloudBaseT', bound='G90CloudBase')
CloudHeaderT = TypeVar('CloudHeaderT', bound='G90CloudHeader')
CloudMessageT = TypeVar('CloudMessageT', bound='G90CloudMessage')

PROTOCOL_VERSION = 1


class G90CloudError(Exception):
    """
    Base exception for G90 cloud protocol errors.
    """


class G90CloudMessageNoMatch(G90CloudError):
    """
    Raised when a message does not match the expected format or type.
    """


class G90CloudMessageInvalid(G90CloudError):
    """
    Raised when a message is invalid or cannot be processed.
    """


@dataclass
class G90CloudMessageContext:  # pylint:disable=too-many-instance-attributes
    """
    Context for G90 cloud messages.

    This class holds information about the local and remote hosts and ports,
    as well as the cloud server and upstream connection details.
    """
    local_host: str
    local_port: int
    remote_host: str
    remote_port: int
    cloud_host: str
    cloud_port: int
    upstream_host: Optional[str]
    upstream_port: Optional[int]
    device_id: Optional[str]


@dataclass
class G90CloudBase:
    """
    Base class for G90 cloud protocol messages.

    Provides methods for encoding and decoding messages to and from their wire
    representation.
    """
    # Format of the binary data representing the message, see `struct` module
    # for the format supported
    _format: ClassVar[str]
    # Context for the message, should be provided when instsantiating the
    # message class
    context: InitVar[G90CloudMessageContext]
    # Stored context
    _context: ClassVar[G90CloudMessageContext]

    @classmethod
    def from_wire(
        cls: Type[CloudBaseT], data: bytes, context: G90CloudMessageContext
    ) -> CloudBaseT:
        """
        Decode a message from its wire representation.

        :param data: The raw bytes of the message.
        :param context: The message context.
        :return: An instance of the message class.
        """
        assert cls.format() is not None
        assert data is not None

        cls._context = context

        try:
            elems = unpack(cls.format(), data[0:cls.size()])
        except StructError as exc:
            raise G90CloudError(
                f"Failed to unpack data: {exc}"
                f", supplied data length: {len(data)}, format: {cls.format()}"
            ) from exc

        try:
            obj = cls(context, *elems)
        except TypeError as exc:
            raise G90CloudError(
                f"Failed to create object: {exc}"
                f", supplied data length: {len(data)}, format: {cls.format()}"
            ) from exc

        return obj

    def to_wire(self) -> bytes:
        """
        Encode the message to its wire representation.

        :return: The raw bytes of the message.
        """
        ret = pack(self.format(), *astuple(self))
        return ret

    @classmethod
    def size(cls) -> int:
        """
        Get the size of the message in bytes.

        :return: The size of the message.
        """
        return calcsize(cls.format())

    @classmethod
    def format(cls) -> str:
        """
        Get the format string for the message.

        :return: The format string.
        """
        return cls._format

    def __str__(self) -> str:
        """
        Get a string representation of the message.

        :return: A string representation of the message.
        """
        return (
            f"{type(self).__name__}("
            f"wire_representation={self.to_wire().hex(' ')}"
        )


@dataclass
class G90CloudHeader(G90CloudBase):
    """
    Header for G90 cloud protocol messages.

    Contains metadata about the message, such as its command, source,
    destination, and payload length.
    """
    _format = '<4Bi'

    command: int
    _source: int
    flag1: int
    _destination: int
    message_length: int

    def __post_init__(self, context: G90CloudMessageContext) -> None:
        """
        Initialize the header and its payload.

        :param context: The message context.
        """
        super().__init__(context)
        self._payload: bytes = bytes()

    @property
    def source(self) -> G90CloudDirection:
        """
        Get the source direction of the message.

        :return: The source direction.
        """
        return G90CloudDirection(self._source)

    @property
    def destination(self) -> G90CloudDirection:
        """
        Get the destination direction of the message.

        :return: The destination direction.
        """
        return G90CloudDirection(self._destination)

    @classmethod
    def from_wire(
        cls: Type[CloudHeaderT], data: bytes, context: G90CloudMessageContext
    ) -> CloudHeaderT:
        """
        Decode a header from its wire representation.

        :param data: The raw bytes of the header.
        :param context: The message context.
        :return: An instance of the header class.
        """
        obj = super().from_wire(data, context)

        message_length = obj.message_length  # pylint:disable=no-member
        if message_length > len(data):
            raise G90CloudError(
                f"Message length of {message_length} specified in header"
                f" exceeds actual data length {len(data)}"
            )

        obj._payload = bytes()  # pylint:disable=protected-access
        if cls.size() < len(data):
            # Payload length in header includes the header size, remove it from
            # resulting payload
            # pylint:disable=no-member,protected-access
            obj._payload = data[cls.size():obj.message_length]
        return obj

    @property
    def payload(self) -> bytes:
        """
        Get the payload of the message.

        :return: The raw bytes of the payload.
        """
        return self._payload

    @property
    def payload_length(self) -> int:
        """
        Get the length of the payload in bytes.

        :return: The payload length.
        """
        return self.message_length - self.size()

    def matches(self, value: G90CloudHeader) -> bool:
        """
        Check if the header matches another header.

        :param value: The header to compare against.
        :return: True if the headers match, False otherwise.
        """
        try:
            return (
                self.command == value.command
                and self._source == value._source
                # pylint:disable=protected-access
                and self._destination == value._destination
            )
        except AttributeError:
            return False

    def __str__(self) -> str:
        """
        Get a string representation of the header.

        :return: A string representation of the header.
        """
        return (
            f"{super().__str__()}"
            f", source={repr(self.source)}"
            f", destination={repr(self.destination)}"
            f", payload({self.payload_length})={self.payload.hex(' ')})"
        )


@dataclass
class G90CloudHeaderVersioned(G90CloudHeader):
    """
    Versioned header for G90 cloud protocol messages.

    Adds version and sequence information to the header.
    """
    _format = '<4BiHH'

    version: int = PROTOCOL_VERSION
    # Sequence will be initialized by :class:`G90CloudMessage` class, has to
    # have a default value as required by `dataclass` (non-default fields
    # cannot follow ones with defaults)
    sequence: int = 0

    def __post_init__(self, context: G90CloudMessageContext) -> None:
        super().__post_init__(context)
        if self.version != PROTOCOL_VERSION:
            raise G90CloudMessageInvalid(
                f'Invalid version in header: {self.version}'
            )

    def __str__(self) -> str:
        """
        Get a string representation of the versioned header.

        :return: A string representation of the versioned header.
        """
        return (
            f"{super().__str__()}"
            f", version={self.version}"
            f", sequence={self.sequence}"
        )


@dataclass
class G90CloudMessage(G90CloudBase):
    """
    Base class for G90 cloud protocol messages with headers.

    Provides methods for encoding and decoding messages with headers, as well
    as handling responses.
    """
    _command: ClassVar[G90CloudCommand]
    _destination: ClassVar[G90CloudDirection]
    _source: ClassVar[G90CloudDirection]
    _responses: ClassVar[List[Type[G90CloudMessage]]] = []
    _header_kls: ClassVar[Type[G90CloudHeader]] = G90CloudHeaderVersioned

    def __post_init__(self, context: G90CloudMessageContext) -> None:
        """
        Initialize the message and its header.

        :param context: The message context.
        """
        self.header = self._header_kls(
            command=self._command, _source=self._source,
            _destination=self._destination, flag1=0,
            message_length=self._header_kls.size() + self.size(),
            context=context
        )

    @classmethod
    def from_wire(
        cls: Type[CloudMessageT], data: bytes, context: G90CloudMessageContext
    ) -> CloudMessageT:
        """
        Decode a message from its wire representation.

        :param data: The raw bytes of the message.
        :param context: The message context.
        :return: An instance of the message class.
        """
        header = cls._header_kls.from_wire(data, context)

        header_matches = cls.header_matches(header)
        if not header_matches:
            raise G90CloudMessageNoMatch('Header does not match')

        try:
            # pylint:disable=no-member
            obj = super().from_wire(header.payload, context)
        except ValueError as exc:
            _LOGGER.error(
                "Failed to create %s from wire: %s",
                cls.__name__, exc
            )
            raise G90CloudMessageNoMatch('Failed to create object') from exc

        obj_matches = cls.matches(obj)
        if not obj_matches:
            raise G90CloudMessageNoMatch('Message does not match')

        obj.header = header
        return obj

    def to_wire(self) -> bytes:
        """
        Encode the message to its wire representation.

        :return: The raw bytes of the message.
        """
        # Reconstruct the wire representation of the block using header plus
        # payload
        ret = self.header.to_wire() + super().to_wire()
        return ret

    def wire_responses(self, context: G90CloudMessageContext) -> List[bytes]:
        """
        Get the wire representations of the responses to this message.

        :param context: The message context.
        :return: A list of raw bytes for the responses.
        """
        result = []
        for idx, response in enumerate(self._responses):
            obj = response(context)
            # Only messages with versioned headers can have sequence numbers
            if isinstance(obj.header, G90CloudHeaderVersioned):
                # Sequence numbers are either 0 for single message responses,
                # or start from 1 if there are multiple ones
                obj.header.sequence = 0
                if len(self._responses) > 1:
                    obj.header.sequence = idx + 1
            _LOGGER.debug(
                "%s: Will send response: %s", type(self).__name__, obj
            )
            result.append(obj.to_wire())
        return result

    @classmethod
    def matches(cls, value: G90CloudMessage) -> bool:
        """
        Check if the message matches another message.

        :param value: The message to compare against.
        :return: True if the messages match, False otherwise.
        """
        try:
            return (
                # pylint:disable=protected-access
                cls._command == value._command
                and cls._source == value._source
                # pylint:disable=protected-access
                and cls._destination == value._destination
            )
        except AttributeError:
            return False

    @classmethod
    def header_matches(cls, value: G90CloudHeader) -> bool:
        """
        Check if the header matches the expected header for this message type.

        :param value: The header to compare against.
        :return: True if the headers match, False otherwise.
        """
        try:
            return (
                cls._command == value.command
                and cls._source == value.source
                and cls._destination == value.destination
            )
        except AttributeError:
            return False

    def __str__(self) -> str:
        """
        Get a string representation of the message.

        :return: A string representation of the message.
        """
        b = [f"{k}={v}" for k, v in asdict(self).items()]
        return (
            f"{type(self).__name__}("
            f"header={str(self.header)}"
            f", wire_representation={self.to_wire().hex(' ')}"
            f", {', '.join(b)})"
        )


class G90CloudStatusChangeReqMessageBase(G90CloudMessage):
    """
    Base class for status change request messages in the G90 cloud protocol.

    Provides methods for handling status change requests and their timestamps.
    """
    _type: ClassVar[G90AlertTypes]

    type: int
    _timestamp: int  # Unix timestamp

    @property
    def timestamp(self) -> datetime:
        """
        Get the timestamp as a datetime object.

        :return: The message timestamp converted to a datetime object with UTC
         timezone.
        """
        return datetime.fromtimestamp(
            self._timestamp, tz=timezone.utc
        )

    @classmethod
    def matches(
        cls, value: G90CloudMessage
    ) -> bool:
        """
        Check if the message matches the expected type and format.

        :param value: The message to compare against.
        :return: True if the messages match, False otherwise.
        """
        try:
            obj = cast(G90CloudStatusChangeReqMessageBase, value)
            return (
                super().matches(value)
                and obj.type == cls._type
            )
        except AttributeError:
            return False
